from rest_framework import filters, pagination


def serializer_factory(endpoint):

    meta_attrs = {
        'model': endpoint.model,
        'fields': endpoint.get_fields_for_serializer()
    }
    meta_parents = (object, )
    if hasattr(endpoint.base_serializer, 'Meta'):
        meta_parents = (endpoint.base_serializer.Meta, ) + meta_parents

    Meta = type('Meta', meta_parents, meta_attrs)

    cls_name = '{}Serializer'.format(endpoint.model.__name__)
    cls_attrs = {
        'Meta': Meta,
    }

    return type(cls_name, (endpoint.base_serializer, ), cls_attrs)


def viewset_factory(endpoint):

    cls_name = '{}ViewSet'.format(endpoint.model.__name__)
    cls_attrs = {
        'serializer_class': endpoint.get_serializer(),
        'queryset': endpoint.model.objects.all(),
        'endpoint': endpoint,
    }

    if endpoint.permission_classes is not None:
        cls_attrs['permission_classes'] = endpoint.permission_classes

    filter_backends = getattr(endpoint.get_base_viewset(), 'filter_backends', ())
    if filter_backends is None:
        filter_backends = []
    else:
        filter_backends = list(filter_backends)

    for filter_type, backend in (
        ('filter_fields', filters.DjangoFilterBackend),
        ('search_fields', filters.SearchFilter),
        ('ordering_fields', filters.OrderingFilter),
    ):

        if getattr(endpoint, filter_type, None) is not None:
            filter_backends.append(backend)
            cls_attrs[filter_type] = getattr(endpoint, filter_type)

    if len(filter_backends) > 0:
        cls_attrs['filter_backends'] = filter_backends

    if endpoint.page_size is not None:
        cls_attrs['page_size'] = endpoint.page_size
        cls_attrs['pagination_class'] = pagination.PageNumberPagination

    return type(cls_name, (endpoint.get_base_viewset(),), cls_attrs)



from rest_framework.metadata import SimpleMetadata, BaseMetadata


class AutoMetadataMixin:

    def determine_metadata(self, request, view):
        from drf_auto_endpoint.app_settings import settings

        metadata = super(AutoMetadataMixin, self).determine_metadata(request, view)
        if not hasattr(view, 'endpoint'):
            serializer_instance = view.serializer_class()
            metadata['fields'] = [
                {
                    'name': field,
                    'label': field.title(),
                    'widget': settings.WIDGET_MAPPING[serializer_instance().fields[field].__class__.__name__],
                }
                for field in view.serializer_class.Meta.fields
                if field != 'id'
            ]
            metadata.update({
                'list_display': ['__str__', ],
                'filter_fields': [],
                'search_fields': [],
                'ordering_fields': [],
                'fieldsets': [{'title': None, 'fields': [
                    field
                    for field in view.serializer_class.Meta.fields
                    if field != 'id' and field != '__str__'
                ]}]
            })
            return metadata

        for prop in ['fields', 'list_display', 'filter_fields', 'search_enabled', 'ordering_fields',
                     'needs', 'fieldsets']:
            metadata[prop] = getattr(view.endpoint, 'get_{}'.format(prop))()

        return metadata


class AutoMetadata(AutoMetadataMixin, SimpleMetadata):
    pass


class MinimalAutMetadata(AutoMetadataMixin, BaseMetadata):
    pass



from rest_framework import serializers, viewsets

from .factories import serializer_factory, viewset_factory


def get_all_field_names(model):
    return [field.name for field in model._meta.get_fields()]


class Endpoint:

    base_serializer = serializers.ModelSerializer
    base_viewset = viewsets.ModelViewSet
    base_readonly_viewset = viewsets.ReadOnlyModelViewSet

    model = None
    fields = None
    serializer = None

    fieldsets = None
    list_display = None

    permission_classes = None
    filter_fields = None
    search_fields = None
    ordering_fields = None
    page_size = None
    viewset = None

    read_only = False
    include_str = True

    def __init__(self, model=None, **kwargs):
        if model is not None:
            self.model = model

        arg_names = ('fields', 'serializer', 'permission_classes', 'filter_fields', 'search_fields',
                     'viewset', 'read_only', 'include_str', 'ordering_fields', 'page_size')
        for arg_name in arg_names:
            setattr(self, arg_name, kwargs.pop(arg_name, getattr(self, arg_name, None)))

        if len(kwargs.keys()) > 0:
            raise Exception('{} got an unexpected keyword argument: "{}"'.format(
                self.__class__.__name__,
                kwargs.keys()[0]
            ))

        if self.serializer is not None:
            assert self.fields is None, 'You cannot specify both fields and serializer'
        else:
            assert self.viewset is not None or self.model is not None, \
                'You need to specify at least a model or a viewset'
            self.get_serializer()

        if self.viewset is not None:
            for attr in ('permission_classes', 'filter_fields', 'search_fields', 'ordering_fields',
                         'page_size'):
                assert getattr(self, attr, None) is None, \
                    'You cannot specify both {} and viewset'.format(attr)
        else:
            self.get_viewset()

        if self.model is None:
            self.model = self.get_serializer().Meta.model

    @property
    def model_name(self):
        return '{}s'.format(self.model._meta.model_name.lower())

    @property
    def application_name(self):
        return self.model._meta.app_label.lower()

    def get_fields_for_serializer(self):

        if self.fields is None:
            if self.fieldsets is None:
                self.fields = tuple(get_all_field_names(self.model))
                if self.include_str:
                    self.fields += ('__str__', )
            else:
                self.fields = [
                    field['name'] if isinstance(field, dict) else field
                    for field in self.fieldsets
                ]

        return self.fields

    def get_serializer(self):

        if self.serializer is None:
            if self.viewset is None:
                self.serializer = serializer_factory(self)
            else:
                self.serializer = self.viewset.serializer_class

        return self.serializer

    def get_base_viewset(self):
        return self.base_viewset if not self.read_only else self.base_readonly_viewset

    def get_viewset(self):

        if self.viewset is None:
            self.viewset = viewset_factory(self)

        return self.viewset

    def get_url(self):

        return '{}/{}'.format(
            self.application_name,
            self.model_name
        )

    def _get_field_dict(self, field):
        from drf_auto_endpoint.app_settings import settings

        serializer_instance = self.get_serializer()()
        name = field['name'] if isinstance(field, dict) else field
        field_instance = serializer_instance.fields[name]
        rv = {
            'name': name,
            'label': name.title() if name != '__str__' else serializer_instance.Meta.model.__name__,
            'widget': settings.WIDGET_MAPPING[field_instance.__class__.__name__],
            'read_only': field_instance.read_only or name == '__str__'
        }

        if isinstance(field, dict):
            rv.update(field)

        return rv

    def get_fields(self):
        return [
            self._get_field_dict(field)
            for field in self.get_fields_for_serializer()
        ]

    def get_fieldsets(self):
        if self.fieldsets is not None:
            return [
                {
                    'title': None,
                    'fields': [{
                        'name': field
                    }
                    if not isinstance(field, dict) else field
                    for field in self.fieldsets ]
                }
            ]
        return [
            {
                'title': None,
                'fields': [{
                    'name': field
                }
                for field in self.get_fields_for_serializer()
                if field != 'id' and field != '__str__' ]
            }
        ]

    def get_list_display(self):
        if self.list_display is None:
            if '__str__' in self.get_fields_for_serializer():
                return ['__str__', ]
            return [self.get_fields()[0]['name']]
        return self.list_display

    def _get_endpoint_list(self, name):
        value = getattr(self, name, None)
        if value is None:
            return []
        return value

    def get_filter_fields(self):
        fields = self._get_endpoint_list('filter_fields')
        return fields

    def get_search_enabled(self):
        fields = self._get_endpoint_list('search_fields')
        if len(fields) > 0:
            return True
        return False

    def get_ordering_fields(self):
        fields = self._get_endpoint_list('ordering_fields')
        return fields

    def get_needs(self):
        related_models = [
            f.model
            if f.model and f.model != self.model
            else f.related_model if f.related_model else None
            for f in self.model._meta.get_fields()
            if ((f.one_to_many or f.one_to_one) and f.auto_created) or (f.many_to_one and f.related_model)
        ]
        return [
            {
                'app': model._meta.app_label,
                'singular': model._meta.model_name.lower(),
                'plural': '{}s'.format(model._meta.model_name.lower()),
            } for model in related_models
        ]




from collections import defaultdict

from django.conf import settings as django_settings


DEFAULT_SETTINGS = {
    'WIDGET_MAPPING': {
        'BooleanField': 'checkbox',
        'NullBooleanField': 'checkbox',
        'IntegerField': 'number',
        'FloatField': 'number',
        'DecimalField': 'number',
        'PrimaryKeyRelatedField': 'foreignkey',
        'ManyRelatedField': 'tomany',
    }
}


class Settings:

    def __init__(self):
        mapping = getattr(django_settings, 'DRF_AUTO_WIDGET_MAPPING', {})
        mapping.update(DEFAULT_SETTINGS['WIDGET_MAPPING'])
        self.WIDGET_MAPPING = defaultdict(lambda: 'text')
        for k, v in mapping.items():
            self.WIDGET_MAPPING[k] = v

    def __getattr__(self, name):
        if name not in DEFAULT_SETTINGS:
            raise AttributeError(name)
        return getattr(django_settings, 'DRF_AUTO_{}'.format(name), DEFAULT_SETTINGS[name])


settings = Settings()



from collections import OrderedDict

from rest_framework.routers import DefaultRouter

from .endpoints import Endpoint


class EndpointRouter(DefaultRouter):

    base_endpoint_class = Endpoint

    def __init__(self, *args, **kwargs):
        self._endpoints = OrderedDict()
        super(EndpointRouter, self).__init__(*args, **kwargs)

    def register(self, model=None, endpoint=None, fields=None, permission_classes=None,
                 serializer=None, filter_fields=None, read_only=False, viewset=None,
                 search_fields=None, ordering_fields=None, page_size=None, **kwargs):

        if endpoint is None:
            endpoint = self.base_endpoint_class(model=model, fields=fields,
                                                permission_classes=permission_classes,
                                                serializer=serializer, filter_fields=filter_fields,
                                                read_only=read_only, viewset=viewset,
                                                search_fields=search_fields,
                                                ordering_fields=ordering_fields)

        url = endpoint.get_url() if 'url' not in kwargs else kwargs.pop('url')
        self._endpoints[url] = endpoint
        super(EndpointRouter, self).register(
            url,
            endpoint.get_viewset(),
            **kwargs
        )

    def get_endpoint(self, url):
        return self._endpoints[url]

    def registerViewSet(self, *args, **kwargs):
        super(EndpointRouter, self).register(*args, **kwargs)


router = EndpointRouter()
